#!/usr/bin/env python

import pytest
from basic_store import BasicStore

from minimalkv.decorator import URLEncodeKeysDecorator
from minimalkv.memory import DictStore


class TestURLEncodeKeysDecorator(BasicStore):
    @pytest.fixture()
    def base_store(self):
        return DictStore()

    @pytest.fixture()
    def store(self, base_store):
        return URLEncodeKeysDecorator(base_store)

    def test_urlencode(self, store):
        store.put("key special:-🍺", b"val1")
        assert store.get("key special:-🍺") == b"val1"

    def test_urlencode_base_store(self, store, base_store, value):
        store.put("abc def/key", value)
        assert "abc+def%2Fkey" in base_store
        assert base_store.get("abc+def%2Fkey") == value
        assert store.get("abc def/key") == value

    # The invalid key is replaced by a valid one after encoding through
    # the decorator...
    test_exception_on_invalid_key_delete = None
    test_exception_on_invalid_key_get_file = None
    test_exception_on_invalid_key_get = None



LAZY_PROPERTY_ATTR_PREFIX = "_lazy_"


def lazy_property(fn):
    """Decorate function to create a property that gets lazy-evaluated.

    On first access, lazy properties are computed and saved
    as instance attribute with the name `'_lazy_' + method_name`
    Any subsequent property access then returns the cached value.
    """
    attr_name = LAZY_PROPERTY_ATTR_PREFIX + fn.__name__

    @property  # type: ignore
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazy_property



import pytest
from basic_store import BasicStore
from conftest import ExtendedKeyspaceTests
from dulwich.repo import Repo
from idgens import HashGen, UUIDGen

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.git import GitCommitStore


class TestGitCommitStore(BasicStore, UUIDGen, HashGen):
    @pytest.fixture(params=[b"master", b"not-master"])
    def branch(self, request):
        return request.param

    @pytest.fixture(params=[b"", b"/sub/subdir"])
    def subdir_name(self, request):
        return request.param

    @pytest.fixture
    def repo_path(self, tmp_path):
        Repo.init_bare(str(tmp_path))
        yield str(tmp_path)

    @pytest.fixture
    def store(self, repo_path, branch, subdir_name):
        return GitCommitStore(repo_path, branch=branch, subdir=subdir_name)

    @pytest.mark.parametrize(
        "subdir_name",
        [b"", b"/", b"subdir", b"sub/subdir/", b"/sub/subdir", b"/sub/subdir/"],
    )
    def test_uses_subdir(self, repo_path, subdir_name, branch):
        store = GitCommitStore(repo_path, branch=branch, subdir=subdir_name)
        # add a key
        store.put("foo", b"bar")

        sdir = subdir_name.decode("ascii").strip("/")

        fn = "foo"

        if sdir:
            fn = sdir + "/" + fn

        repo = Repo(repo_path)
        commit = repo[repo.refs[b"refs/heads/" + branch]]
        tree = repo[commit.tree]
        _, blob_id = tree.lookup_path(repo.__getitem__, fn.encode("ascii"))

        assert repo[blob_id].data == b"bar"

        # add a second key, resulting in a new commit and two keys available
        store.put("foo2", b"bar2")

        fn = "foo"
        if sdir:
            fn = sdir + "/" + fn

        repo = Repo(repo_path)
        commit = repo[repo.refs[b"refs/heads/" + branch]]
        tree = repo[commit.tree]
        _, blob_id = tree.lookup_path(repo.__getitem__, fn.encode("ascii"))
        assert repo[blob_id].data == b"bar"

        fn2 = "foo2"
        if sdir:
            fn2 = sdir + "/" + fn2
        _, blob_id = tree.lookup_path(repo.__getitem__, fn2.encode("ascii"))
        assert repo[blob_id].data == b"bar2"


class TestExtendedKeyspaceGitStore(TestGitCommitStore, ExtendedKeyspaceTests):
    @pytest.fixture
    def store(self, repo_path, branch, subdir_name):
        class ExtendedKeyspaceStore(ExtendedKeyspaceMixin, GitCommitStore):
            pass

        return ExtendedKeyspaceStore(repo_path, branch=branch, subdir=subdir_name)



import re

FOREVER = "forever"
NOT_SET = "not_set"

VALID_NON_NUM = r"""\`\!"#$%&'()+,-.<=>?@[]^_{}~"""
VALID_KEY_REGEXP = "^[%s0-9a-zA-Z]+$" % re.escape(VALID_NON_NUM)
"""This regular expression tests if a key is valid. Allowed are all
alphanumeric characters, as well as ``!"`#$%&'()+,-.<=>?@[]^_{}~``."""

VALID_KEY_RE = re.compile(VALID_KEY_REGEXP)
"""A compiled version of :data:`~minimalkv._constants.VALID_KEY_REGEXP`."""

VALID_NON_NUM_EXTENDED = VALID_NON_NUM + r"/ "
VALID_KEY_REGEXP_EXTENDED = "^[%s0-9a-zA-Z]+$" % re.escape(VALID_NON_NUM_EXTENDED)
"""This regular expression tests if a key is valid when the extended keyspace mixin is used.
Allowed are all alphanumeric characters, as well as ``!"`#$%&'()+,-.<=>?@[]^_{}~/``. and spaces"""
VALID_KEY_RE_EXTENDED = re.compile(VALID_KEY_REGEXP_EXTENDED)
"""A compiled version of :data:`~minimalkv._constants.VALID_KEY_REGEXP_EXTENDED`."""



import pickle
from base64 import b64encode
from configparser import ConfigParser
from uuid import uuid4 as uuid

import pytest
from basic_store import BasicStore, OpenSeekTellStore
from conftest import ExtendedKeyspaceTests

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.net.azurestore import AzureBlockBlobStore

asb = pytest.importorskip("azure.storage.blob")


def get_azure_conn_string():
    cfg_fn = "azure_credentials.ini"
    parser = ConfigParser(
        {
            "protocol": "https",
            "endpoint": "",
            "account_name": "",
        }
    )
    result = parser.read(cfg_fn)
    if not result:
        pytest.skip(f"file {cfg_fn} not found")

    for section in parser.sections():
        account_name = parser.get(section, "account_name")
        if not account_name:
            pytest.skip(f"no 'account_name' found in file {cfg_fn}")

        account_key = parser.get(section, "account_key")
        protocol = parser.get(section, "protocol")
        endpoint = parser.get(section, "endpoint")
        conn_string = "DefaultEndpointsProtocol={};AccountName={};AccountKey={}".format(
            protocol, account_name, account_key
        )
        if endpoint:
            conn_string += f";BlobEndpoint={endpoint}"
        return conn_string


def _delete_container(conn_string, container):
    try:
        # for azure-storage-blob>=12:
        from azure.core.exceptions import AzureError
        from azure.storage.blob import BlobServiceClient

        s = BlobServiceClient.from_connection_string(conn_string)
        try:
            s.delete_container(container)
        except AzureError as ex:
            # ignore the ContainerNotFound error:
            if ex.error_code != "ContainerNotFound":
                raise
    except ImportError:
        # for azure-storage-blob<12
        from azure.storage.blob import BlockBlobService

        s = BlockBlobService(connection_string=conn_string)
        s.delete_container(container)


class TestAzureStorage(BasicStore, OpenSeekTellStore):
    @pytest.fixture
    def store(self):
        container = str(uuid())
        conn_string = get_azure_conn_string()
        yield AzureBlockBlobStore(
            conn_string=conn_string, container=container, public=False
        )
        _delete_container(conn_string, container)


class TestExtendedKeysAzureStorage(TestAzureStorage, ExtendedKeyspaceTests):
    @pytest.fixture
    def store(self):
        azure_storage_blob_major_version = int(asb.__version__.split(".", 1)[0])
        conn_string = get_azure_conn_string()
        use_azurite = "http://127.0.0.1:10000/devstoreaccount1" in conn_string
        if use_azurite and azure_storage_blob_major_version < 12:
            pytest.skip("Compatibility issues with azurite and azure-storage-blob<12")
        container = str(uuid())

        class ExtendedKeysStore(ExtendedKeyspaceMixin, AzureBlockBlobStore):
            pass

        yield ExtendedKeysStore(
            conn_string=conn_string, container=container, public=False
        )
        _delete_container(conn_string, container)


def test_azure_setgetstate():
    container = str(uuid())
    conn_string = get_azure_conn_string()
    store = AzureBlockBlobStore(conn_string=conn_string, container=container)
    store.put("key1", b"value1")

    buf = pickle.dumps(store, protocol=2)
    store = pickle.loads(buf)

    assert store.get("key1") == b"value1"
    _delete_container(conn_string, container)


def test_azure_store_attributes():
    abbs = AzureBlockBlobStore(
        "CONN_STR", "CONTAINER", max_connections=42, checksum=True
    )
    assert abbs.conn_string == "CONN_STR"
    assert abbs.container == "CONTAINER"
    assert abbs.public is False
    assert abbs.create_if_missing is True
    assert abbs.max_connections == 42
    assert abbs.checksum is True

    abbs2 = pickle.loads(pickle.dumps(abbs))
    assert abbs2.conn_string == "CONN_STR"
    assert abbs2.container == "CONTAINER"
    assert abbs2.public is False
    assert abbs2.create_if_missing is True
    assert abbs2.max_connections == 42
    assert abbs2.checksum is True


def test_azure_special_args():
    # For azure-storage-blob 12,
    # test that the special arguments `max_block_size` and
    # `max_single_put_size` propagate to the constructed ContainerClient
    conn_string = get_azure_conn_string()
    MBS = 983645
    MSP = 756235
    abbs = AzureBlockBlobStore(
        conn_string=conn_string,
        container="container-unused",
        max_block_size=MBS,
        max_single_put_size=MSP,
        create_if_missing=False,
    )
    if hasattr(abbs, "blob_container_client"):
        cfg = abbs.blob_container_client._config  # type: ignore
        assert cfg.max_single_put_size == MSP
        assert cfg.max_block_size == MBS


class TestAzureExceptionHandling:
    def test_missing_container(self):
        container = str(uuid())
        conn_string = get_azure_conn_string()
        store = AzureBlockBlobStore(
            conn_string=conn_string, container=container, create_if_missing=False
        )
        with pytest.raises(IOError) as exc:
            store.keys()
        assert "The specified container does not exist." in str(exc.value)

    def test_wrong_endpoint(self):
        container = str(uuid())
        conn_string = get_azure_conn_string()
        conn_settings = dict([s.split("=", 1) for s in conn_string.split(";") if s])
        conn_settings["BlobEndpoint"] = "https://host-does-not-exist/"
        conn_string = ";".join(f"{key}={value}" for key, value in conn_settings.items())
        store = AzureBlockBlobStore(
            conn_string=conn_string, container=container, create_if_missing=False
        )
        if hasattr(store, "block_blob_service"):
            from azure.storage.common.retry import ExponentialRetry

            store.block_blob_service.retry = ExponentialRetry(max_attempts=0).retry
        else:
            store.blob_container_client._config.retry_policy.total_retries = 0  # type: ignore

        with pytest.raises(IOError) as exc:
            store.put("key", b"data")
        assert "connect" in str(exc.value)

    def test_wrong_credentials(self):
        container = str(uuid())
        conn_string = (
            "DefaultEndpointsProtocol=https;AccountName={};AccountKey={}".format(
                "testaccount", "wrongsecret"
            )
        )
        store = AzureBlockBlobStore(
            conn_string=conn_string, container=container, create_if_missing=False
        )

        if hasattr(store, "block_blob_service"):
            from azure.storage.common.retry import ExponentialRetry

            store.block_blob_service.retry = ExponentialRetry(max_attempts=0).retry
        else:
            store.blob_container_client._config.retry_policy.total_retries = 0  # type: ignore

        with pytest.raises(IOError) as exc:
            store.put("key", b"data")
        assert "Incorrect padding" in str(exc.value)


class TestChecksum:
    CONTENT = b"\1\2\3\4"
    EXPECTED_CHECKSUM = "CNbAWiFRKnmh3+udKo8mLw=="
    KEY = "testkey"

    @pytest.fixture
    def store(self):
        container = str(uuid())
        conn_string = get_azure_conn_string()

        yield AzureBlockBlobStore(
            conn_string=conn_string,
            container=container,
            public=False,
            checksum=True,
        )
        _delete_container(conn_string, container)

    def _checksum(self, store):
        # request the md5 checksum from azure and return the b64 encoded value
        if hasattr(store, "block_blob_service"):
            return store.block_blob_service.get_blob_properties(
                store.container,
                self.KEY,
            ).properties.content_settings.content_md5
        else:
            checksum_bytes = (
                store.blob_container_client.get_blob_client(self.KEY)
                .get_blob_properties()
                .content_settings.content_md5
            )
            return b64encode(checksum_bytes).decode()

    def test_checksum_put(self, store):
        store.put(self.KEY, self.CONTENT)
        assert self._checksum(store) == self.EXPECTED_CHECKSUM
        assert store.get(self.KEY) == self.CONTENT

    def test_checksum_put_file(self, store, tmpdir):
        file_ = tmpdir.join("my_file")
        file_.write(self.CONTENT)
        store.put_file(self.KEY, file_.open("rb"))
        assert self._checksum(store) == self.EXPECTED_CHECKSUM
        assert store.get(self.KEY) == self.CONTENT

    def test_checksum_put_filename(self, store, tmpdir):
        file_ = tmpdir.join("my_file")
        file_.write(self.CONTENT)
        store.put_file(self.KEY, str(file_))
        assert self._checksum(store) == self.EXPECTED_CHECKSUM
        assert store.get(self.KEY) == self.CONTENT



#!/usr/bin/env python

from uuid import uuid4 as uuid

import pytest

pymongo = pytest.importorskip("pymongo")

from basic_store import BasicStore
from conftest import ExtendedKeyspaceTests

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.db.mongo import MongoStore


class TestMongoDB(BasicStore):
    @pytest.fixture
    def db_name(self):
        return f"_minimalkv_test_{uuid()}"

    @pytest.fixture
    def store(self, db_name):
        try:
            conn = pymongo.MongoClient()
        except pymongo.errors.ConnectionFailure:
            pytest.skip("could not connect to mongodb")
        yield MongoStore(conn[db_name], "minimalkv-tests")
        conn.drop_database(db_name)


class TestExtendedKeyspaceDictStore(TestMongoDB, ExtendedKeyspaceTests):
    @pytest.fixture
    def store(self, db_name):
        class ExtendedKeyspaceStore(ExtendedKeyspaceMixin, MongoStore):
            pass

        try:
            conn = pymongo.MongoClient()
        except pymongo.errors.ConnectionFailure:
            pytest.skip("could not connect to mongodb")
        yield ExtendedKeyspaceStore(conn[db_name], "minimalkv-tests")
        conn.drop_database(db_name)



#!/usr/bin/env python
import pytest
from basic_store import BasicStore
from conftest import ExtendedKeyspaceTests
from idgens import HashGen, UUIDGen
from test_hmac import HMACDec

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.memory import DictStore


class TestDictStore(BasicStore, UUIDGen, HashGen, HMACDec):
    @pytest.fixture
    def store(self):
        return DictStore()


class TestExtendedKeyspaceDictStore(TestDictStore, ExtendedKeyspaceTests):
    @pytest.fixture
    def store(self):
        class ExtendedKeyspaceStore(ExtendedKeyspaceMixin, DictStore):
            pass

        return ExtendedKeyspaceStore()



import pytest
from basic_store import BasicStore

from minimalkv.cache import CacheDecorator
from minimalkv.memory import DictStore


class TestCache(BasicStore):
    # FIXME: this could use some extra combinations tested
    @pytest.fixture
    def front_store(self):
        return DictStore()

    @pytest.fixture
    def backing_store(self):
        return DictStore()

    @pytest.fixture
    def store(self, front_store, backing_store):
        return CacheDecorator(front_store, backing_store)

    def test_works_when_cache_loses_key(self, store, front_store, key, value):
        store.put(key, value)

        assert store.get(key) == value

        front_store.delete(key)

        assert store.get(key) == value



try:
    from azure.storage.blob import BlockBlobService  # noqa: F401

    from ._azurestore_old import AzureBlockBlobStore
except ImportError:
    from ._azurestore_new import AzureBlockBlobStore  # type: ignore

__all__ = ["AzureBlockBlobStore"]



import os
import re
import tempfile
import uuid

import pytest

from minimalkv.idgen import HashDecorator, UUIDDecorator

UUID_REGEXP = re.compile(
    r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
)


class IDGen:
    @pytest.fixture(
        params=[
            "constant",
            "foo{}bar",
            "{}.jpeg",
            "prefix-{}.hello",
            "justprefix{}",
        ]
    )
    def idgen_template(self, request):
        return request.param


class UUIDGen(IDGen):
    @pytest.fixture
    def uuidstore(self, store):
        return UUIDDecorator(store)

    @pytest.fixture
    def templated_uuidstore(self, store, idgen_template):
        return UUIDDecorator(store, idgen_template)

    def test_put_generates_uuid_form(self, uuidstore, value):
        key = uuidstore.put(None, value)
        assert UUID_REGEXP.match(key)

    def test_put_file_generates_uuid_form(self, uuidstore):
        key = uuidstore.put_file(None, open("/dev/null", "rb"))
        assert UUID_REGEXP.match(key)

        tmpfile = tempfile.NamedTemporaryFile(delete=False)
        try:
            tmpfile.close()
            key2 = uuidstore.put_file(None, tmpfile.name)
            assert UUID_REGEXP.match(key2)
        finally:
            if os.path.exists(tmpfile.name):
                os.unlink(tmpfile.name)

    def test_put_generates_valid_uuid(self, uuidstore, value):
        key = uuidstore.put(None, value)
        uuid.UUID(hex=key)

    def test_put_file_generates_valid_uuid(self, uuidstore):
        key = uuidstore.put_file(None, open("/dev/null", "rb"))
        uuid.UUID(hex=key)

        tmpfile = tempfile.NamedTemporaryFile(delete=False)
        try:
            tmpfile.close()
            key2 = uuidstore.put_file(None, tmpfile.name)
            uuid.UUID(hex=key2)
        finally:
            if os.path.exists(tmpfile.name):
                os.unlink(tmpfile.name)

    def test_templates_work_uuid(self, templated_uuidstore, value, idgen_template):
        key = templated_uuidstore.put(None, value)

        # should not be a valid UUID
        assert not UUID_REGEXP.match(key)


class HashGen(IDGen):
    @pytest.fixture
    def hashstore(self, store, hashfunc):
        return HashDecorator(store, hashfunc)

    @pytest.fixture
    def templated_hashstore(self, store, hashfunc, idgen_template):
        return HashDecorator(store, hashfunc, idgen_template)

    @pytest.fixture
    def validate_hash(self, hashfunc):
        hash_regexp = re.compile(
            r"^[0-9a-f]{{{}}}$".format(
                hashfunc().digest_size * 2,
            )
        )

        return hash_regexp.match

    @pytest.fixture
    def value_hash(self, hashfunc, value):
        return hashfunc(value).hexdigest()

    def test_put_generates_valid_form(self, hashstore, validate_hash, value):
        key = hashstore.put(None, value)
        assert validate_hash(key)

    def test_put_file_generates_valid_form(self, hashstore, validate_hash):
        key = hashstore.put_file(None, open("/dev/null", "rb"))
        assert validate_hash(key)

        # this is not correct according to our interface
        # /dev/null cannot be claimed by the hashstore
        # key2 = hashstore.put_file(None, '/dev/null')
        # assert validate_hash(key2)

    def test_put_generates_correct_hash(self, hashstore, value_hash, value):
        key = hashstore.put(None, value)

        assert value_hash == key
        assert isinstance(key, str)

    def test_put_file_generates_correct_hash(self, hashstore, value_hash, value):
        tmpfile = tempfile.NamedTemporaryFile(delete=False)
        try:
            tmpfile.write(value)
            tmpfile.close()
            with open(tmpfile.name, "rb") as f:
                key = hashstore.put_file(None, f)
            assert key == value_hash

            key2 = hashstore.put_file(None, tmpfile.name)
            assert key2 == value_hash
        finally:
            if os.path.exists(tmpfile.name):
                os.unlink(tmpfile.name)

    def test_templates_work(
        self, templated_hashstore, value, idgen_template, value_hash
    ):
        key = templated_hashstore.put(None, value)

        assert idgen_template.format(value_hash) == key



import pytest

storage = pytest.importorskip("google.cloud.storage")

import os
import pickle
import time
from configparser import ConfigParser
from typing import Optional
from uuid import uuid4

from basic_store import BasicStore, OpenSeekTellStore
from conftest import ExtendedKeyspaceTests
from google.api_core.exceptions import NotFound
from google.auth.credentials import AnonymousCredentials
from google.cloud.exceptions import MethodNotAllowed
from google.cloud.storage import Bucket, Client

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.net.gcstore import GoogleCloudStore


@pytest.fixture(scope="module")
def gc_credentials():
    parser = ConfigParser()
    parser.read("google_cloud_credentials.ini")
    credentials_path = parser.get(
        "google-cloud-tests", "credentials_json_path", fallback=None
    )
    emulator_endpoint = parser.get(
        "google-cloud-tests", "emulator_endpoint", fallback=None
    )

    assert (
        credentials_path or emulator_endpoint
    ), "Either set endpoint (for gc emulation) or credentials_json_path (for actual gc)"

    if emulator_endpoint:
        # google's client library looks for this env var
        # if we didn't set it it would use the standard endpoint
        # at https://storage.googleapis.com
        os.environ["STORAGE_EMULATOR_HOST"] = emulator_endpoint
        credentials = AnonymousCredentials()
    else:
        # if no endpoint was defined we're running against actual GC and need credentials
        credentials = credentials_path
    yield credentials
    # unset the env var
    if emulator_endpoint:
        del os.environ["STORAGE_EMULATOR_HOST"]


def try_delete_bucket(bucket):
    # normally here we should delete the bucket
    # however the emulator (fake-gcs-server) doesn't currently support bucket deletion.
    # see: https://github.com/fsouza/fake-gcs-server/issues/214
    # So we empty the bucket and then try to delete it
    for blob in bucket.list_blobs():
        blob.delete()
    try:
        bucket.delete()
    except MethodNotAllowed as err:
        # closely match error thrown by fake gcs so we notice if something changes
        assert err.code == 405
        assert err.message.endswith("unknown error")


def get_bucket_from_store(gcstore: GoogleCloudStore) -> Bucket:
    client = get_client_from_store(gcstore)
    if gcstore.create_if_missing and not client.lookup_bucket(gcstore.bucket_name):
        bucket = client.create_bucket(
            bucket_or_name=gcstore.bucket_name,
            location=gcstore.bucket_creation_location,
        )
    else:
        # will raise an error if bucket not found
        bucket = client.get_bucket(gcstore.bucket_name)

    return bucket


def get_client_from_store(gcstore: GoogleCloudStore) -> Client:
    if type(gcstore._credentials) == str:
        client = Client.from_service_account_json(gcstore._credentials)
    else:
        client = Client(credentials=gcstore._credentials, project=gcstore.project_name)

    return client


@pytest.fixture(scope="module")
def dirty_store(gc_credentials):
    uuid = str(uuid4())
    # if we have a credentials.json that specifies the project name, else we pick one
    if type(gc_credentials) == AnonymousCredentials:
        project_name: Optional[str] = "testing"
    else:
        project_name = None
    store = GoogleCloudStore(
        credentials=gc_credentials, bucket_name=uuid, project=project_name
    )
    yield store
    try_delete_bucket(get_bucket_from_store(store))


@pytest.fixture(scope="function")
def store(dirty_store):
    for blob in get_bucket_from_store(dirty_store).list_blobs():
        blob.delete()

    dirty_store._fs.invalidate_cache()

    # Google Storage doesn't like getting hit with heavy CRUD on a newly
    # create bucket. Therefore we introduce an artificial timeout
    if not os.environ.get("STORAGE_EMULATOR_HOST", None):
        time.sleep(0.3)
    return dirty_store


class TestGoogleCloudStore(OpenSeekTellStore, BasicStore):
    pass


def test_gcstore_pickling(store):
    store.put("key1", b"value1")

    buf = pickle.dumps(store)
    store = pickle.loads(buf)

    assert store.get("key1") == b"value1"


def test_gcstore_pickling_attrs():
    store = GoogleCloudStore(
        credentials="path_to_json",
        bucket_name="test_bucket",
        create_if_missing=False,
        bucket_creation_location="US-CENTRAL1",
        project="sample_project",
    )

    buf = pickle.dumps(store)
    store = pickle.loads(buf)

    assert store.bucket_name == "test_bucket"
    assert not store.create_if_missing
    assert store.bucket_creation_location == "US-CENTRAL1"
    assert store.project_name == "sample_project"


class TestExtendedKeysGCStore(TestGoogleCloudStore, ExtendedKeyspaceTests):
    @pytest.fixture(scope="class")
    def dirty_store(self, gc_credentials):
        # This overwrites the module-level fixture and returns an ExtendedKeysStore
        # instead of the regular one
        uuid = str(uuid4())
        # if we have a credentials.json that specifies the project name, else we pick one
        if type(gc_credentials) == AnonymousCredentials:
            project_name: Optional[str] = "testing"
        else:
            project_name = None

        class ExtendedKeysStore(ExtendedKeyspaceMixin, GoogleCloudStore):
            pass

        store = ExtendedKeysStore(
            credentials=gc_credentials, bucket_name=uuid, project=project_name
        )
        yield store
        try_delete_bucket(get_bucket_from_store(store))

    @pytest.fixture(scope="function")
    def store(self, dirty_store):
        for blob in get_bucket_from_store(dirty_store).list_blobs():
            blob.delete()
        # Invalidate fsspec cache
        dirty_store._fs.invalidate_cache()
        if not os.environ.get("STORAGE_EMULATOR_HOST", None):
            time.sleep(0.2)
        return dirty_store


class TestGCExceptions:
    def test_nonexisting_bucket(self, gc_credentials):
        store = GoogleCloudStore(
            credentials=gc_credentials,
            bucket_name="thisbucketdoesntexist123123",
            create_if_missing=False,
        )
        with pytest.raises(NotFound):
            store.get("key")



from io import BytesIO
from typing import IO, Iterator

from sqlalchemy import Column, LargeBinary, String, Table, exists, select

from minimalkv import CopyMixin, KeyValueStore


class SQLAlchemyStore(KeyValueStore, CopyMixin):  # noqa D
    def __init__(self, bind, metadata, tablename):
        self.bind = bind

        self.table = Table(
            tablename,
            metadata,
            # 250 characters is the maximum key length that we guarantee can be
            # handled by any kind of backend
            Column("key", String(250), primary_key=True),
            Column("value", LargeBinary, nullable=False),
        )

    def _has_key(self, key: str) -> bool:
        return self.bind.execute(
            select([exists().where(self.table.c.key == key)])
        ).scalar()

    def _delete(self, key: str) -> None:
        self.bind.execute(self.table.delete(self.table.c.key == key))

    def _get(self, key: str) -> bytes:
        rv = self.bind.execute(
            select([self.table.c.value], self.table.c.key == key).limit(1)
        ).scalar()

        if not rv:
            raise KeyError(key)

        return rv

    def _open(self, key: str) -> IO:
        return BytesIO(self._get(key))

    def _copy(self, source: str, dest: str):
        con = self.bind.connect()
        with con.begin():
            data = self.bind.execute(
                select([self.table.c.value], self.table.c.key == source).limit(1)
            ).scalar()
            if not data:
                raise KeyError(source)

            # delete the potential existing previous key
            con.execute(self.table.delete(self.table.c.key == dest))
            con.execute(
                self.table.insert(
                    {
                        "key": dest,
                        "value": data,
                    }
                )
            )
        con.close()
        return dest

    def _put(self, key: str, data: bytes) -> str:
        con = self.bind.connect()
        with con.begin():
            # delete the old
            con.execute(self.table.delete(self.table.c.key == key))

            # insert new
            con.execute(self.table.insert({"key": key, "value": data}))

            # commit happens here

        con.close()
        return key

    def _put_file(self, key: str, file: IO) -> str:
        return self._put(key, file.read())

    def iter_keys(self, prefix: str = "") -> Iterator[str]:  # noqa D
        query = select([self.table.c.key])
        if prefix != "":
            query = query.where(self.table.c.key.like(prefix + "%"))
        return map(lambda v: str(v[0]), self.bind.execute(query))



from typing import Iterable
from urllib.parse import quote_plus, unquote_plus

from minimalkv._key_value_store import KeyValueStore


class StoreDecorator:
    """Base class for store decorators.

    The default implementation will use :func:`getattr` to pass through all
    attribute/method requests to an underlying object stored as
    :attr:`_dstore`. It will also pass through the :attr:`__getattr__` and
    :attr:`__contains__` python special methods.

    Attributes
    ----------
    _dstore: KeyValueStore
        Store.

    Parameters
    ----------
    store: KeyValueStore
        Store.
    """

    def __init__(self, store: KeyValueStore):
        self._dstore = store

    def __getattr__(self, attr):  # noqa D
        # Use object.__getattritbute__ as getattr  would make a recursive call to
        # StoreDecorator.__getattr__.
        store = object.__getattribute__(self, "_dstore")
        return getattr(store, attr)

    def __contains__(self, key: str) -> bool:  # noqa D
        return self._dstore.__contains__(key)

    def __iter__(self) -> Iterable[str]:  # noqa D
        return self._dstore.__iter__()


class KeyTransformingDecorator(StoreDecorator):  # noqa D
    # TODO Document KeyTransformingDecorator.
    # currently undocumented (== not advertised as a feature)
    def _map_key(self, key: str) -> str:
        return key

    def _map_key_prefix(self, key_prefix: str) -> str:
        return key_prefix

    def _unmap_key(self, key: str) -> str:
        return key

    def _filter(self, key: str) -> bool:
        return True

    def __contains__(self, key: str) -> bool:  # noqa D
        return self._map_key(key) in self._dstore

    def __iter__(self) -> Iterable[str]:  # noqa D
        return self.iter_keys()

    def delete(self, key: str):  # noqa D
        return self._dstore.delete(self._map_key(key))

    def get(self, key, *args, **kwargs):  # noqa D
        return self._dstore.get(self._map_key(key), *args, **kwargs)  # type: ignore

    def get_file(self, key: str, *args, **kwargs):  # noqa D
        return self._dstore.get_file(self._map_key(key), *args, **kwargs)

    def iter_keys(self, prefix: str = "") -> Iterable[str]:  # noqa D
        return (
            self._unmap_key(k)
            for k in self._dstore.iter_keys(self._map_key_prefix(prefix))
            if self._filter(k)
        )

    def iter_prefixes(  # noqa D
        self, delimiter: str, prefix: str = ""
    ) -> Iterable[str]:
        dlen = len(delimiter)
        plen = len(prefix)
        memory = set()

        for k in self.iter_keys(prefix):
            pos = k.find(delimiter, plen)
            if pos >= 0:
                k = k[: pos + dlen]

            if k not in memory:
                yield k
                memory.add(k)

    def keys(self, prefix: str = ""):  # noqa D
        return list(self.iter_keys(prefix))

    def open(self, key: str):  # noqa D
        return self._dstore.open(self._map_key(key))

    def put(self, key: str, *args, **kwargs):  # noqa D
        return self._unmap_key(self._dstore.put(self._map_key(key), *args, **kwargs))

    def put_file(self, key: str, *args, **kwargs):  # noqa D
        return self._unmap_key(
            self._dstore.put_file(self._map_key(key), *args, **kwargs)
        )

    # support for UrlMixin
    def url_for(self, key: str, *args, **kwargs) -> str:  # noqa D
        return self._dstore.url_for(self._map_key(key), *args, **kwargs)  # type: ignore

    # support for CopyMixin
    def copy(self, source: str, dest: str):  # noqa D
        return self._dstore.copy(self._map_key(source), self._map_key(dest))  # type: ignore


class PrefixDecorator(KeyTransformingDecorator):
    """
    Prefixes any key with a string before passing it on the decorated store.

    Automatically strips the prefix upon key retrieval.

    Parameters
    ----------
    store : KeyValueStore
        The store to pass keys on to.
    prefix : str
        Prefix to add.

    """

    def __init__(self, prefix: str, store: KeyValueStore):
        super().__init__(store)
        self.prefix = prefix

    def _filter(self, key: str) -> bool:
        return key.startswith(self.prefix)

    def _map_key(self, key: str) -> str:
        self._check_valid_key(key)
        return self.prefix + key

    def _map_key_prefix(self, key_prefix: str) -> str:
        return self.prefix + key_prefix

    def _unmap_key(self, key: str) -> str:
        assert key.startswith(self.prefix)

        return key[len(self.prefix) :]


class URLEncodeKeysDecorator(KeyTransformingDecorator):
    """URL-encodes keys before passing them on to the underlying store."""

    def _map_key(self, key: str) -> str:  # noqa D
        if not isinstance(key, str):
            raise ValueError("%r is not a unicode string" % key)
        quoted = quote_plus(key.encode("utf-8"))
        if isinstance(quoted, bytes):
            quoted = quoted.decode("utf-8")
        return quoted

    def _map_key_prefix(self, key_prefix: str) -> str:  # noqa D
        return self._map_key(key_prefix)

    def _unmap_key(self, key: str) -> str:  # noqa D
        return unquote_plus(key)


class ReadOnlyDecorator(StoreDecorator):
    """A read-only view of an underlying minimalkv store.

    Provides only access to the following methods/attributes of the underlying store:
    ``get``, ``iter_keys``, ``keys``, ``open``, ``get_file`` and ``__contains__``.
    Accessing any other method will raise ``AttributeError``.

    Note that the original store for read / write can still be accessed, so using this
    class as a wrapper only provides protection against bugs and other kinds of
    unintentional writes; it is not meant to be a real security measure.

    """

    def __getattr__(self, attr):  # noqa D
        if attr in ("get", "iter_keys", "keys", "open", "get_file"):
            return super().__getattr__(attr)
        else:
            raise AttributeError



import os
import tempfile
import time
from io import BytesIO

import pytest
from pyarrow.parquet import ParquetFile

from minimalkv import CopyMixin
from minimalkv.crypt import HMACDecorator
from minimalkv.decorator import PrefixDecorator
from minimalkv.idgen import HashDecorator, UUIDDecorator
from minimalkv.net.gcstore import GoogleCloudStore


def is_emulated_gcstore_test(store):
    return isinstance(store, GoogleCloudStore) and "localhost" in os.environ.get(
        "STORAGE_EMULATOR_HOST", ""
    )


class BasicStore:
    def test_store(self, store, key, value):
        new_key = store.put(key, value)
        assert key == new_key

    def test_unicode_store(self, store, key, unicode_value):
        with pytest.raises(IOError):
            store.put(key, unicode_value)

    def test_store_and_retrieve(self, store, key, value):
        store.put(key, value)
        assert store.get(key) == value

    def test_store_and_retrieve_filelike(self, store, key, value):
        store.put_file(key, BytesIO(value))
        assert store.get(key) == value

    def test_store_and_retrieve_overwrite(self, store, key, value, value2):
        store.put_file(key, BytesIO(value))
        assert store.get(key) == value

        store.put(key, value2)
        assert store.get(key) == value2

    def test_store_and_open(self, store, key, value):
        store.put_file(key, BytesIO(value))
        assert store.open(key).read() == value

    def test_store_and_copy(self, store, key, key2, value):
        if not isinstance(store, CopyMixin):
            pytest.skip()
        store.put(key, value)
        assert store.get(key) == value
        store.copy(key, key2)
        assert store.get(key) == value
        assert store.get(key2) == value

    def test_store_and_copy_overwrite(self, store, key, key2, value, value2):
        if not isinstance(store, CopyMixin):
            pytest.skip()
        store.put(key, value)
        store.put(key2, value2)
        assert store.get(key) == value
        assert store.get(key2) == value2
        store.copy(key, key2)
        assert store.get(key) == value
        assert store.get(key2) == value

    def test_open_incremental_read(self, store, key, long_value):
        store.put_file(key, BytesIO(long_value))
        ok = store.open(key)
        assert long_value[:3] == ok.read(3)
        assert long_value[3:5] == ok.read(2)
        assert long_value[5:8] == ok.read(3)

    def test_key_error_on_nonexistant_get(self, store, key):
        with pytest.raises(KeyError):
            store.get(key)

    def test_key_error_on_nonexistant_copy(self, store, key, key2):
        if not isinstance(store, CopyMixin):
            pytest.skip()
        with pytest.raises(KeyError):
            store.copy(key, key2)

    def test_key_error_on_nonexistant_open(self, store, key):
        with pytest.raises(KeyError):
            store.open(key)

    def test_key_error_on_nonexistant_get_file(self, store, key):
        with pytest.raises(KeyError):
            store.get_file(key, BytesIO())

    def test_key_error_on_nonexistant_get_filename(self, store, key):
        with pytest.raises(KeyError):
            store.get_file(key, "/dev/null")

    def test_exception_on_invalid_key_get(self, store, invalid_key):
        with pytest.raises(ValueError):
            store.get(invalid_key)

    def test_exception_on_invalid_key_copy(self, store, invalid_key, key):
        if not isinstance(store, CopyMixin):
            pytest.skip()
        with pytest.raises(ValueError):
            store.copy(invalid_key, key)
        with pytest.raises(ValueError):
            store.copy(key, invalid_key)

    def test_exception_on_invalid_key_get_file(self, store, invalid_key):
        with pytest.raises(ValueError):
            store.get_file(invalid_key, "/dev/null")

    def test_exception_on_invalid_key_delete(self, store, invalid_key):
        with pytest.raises(ValueError):
            store.delete(invalid_key)

    def test_put_file(self, store, key, value, request):
        if is_emulated_gcstore_test(store):
            mark = pytest.mark.xfail(
                reason="Triggers resumable upload, which isn't currently supported by the GC Emulator"
            )
            request.node.add_marker(mark)

        tmp = tempfile.NamedTemporaryFile(delete=False)
        try:
            tmp.write(value)
            tmp.close()

            store.put_file(key, tmp.name)

            assert store.get(key) == value
        finally:
            if os.path.exists(tmp.name):
                os.unlink(tmp.name)

    def test_put_opened_file(self, store, key, value, request):
        if is_emulated_gcstore_test(store):
            mark = pytest.mark.xfail(
                reason="Triggers resumable upload, which isn't currently supported by the GC Emulator"
            )
            request.node.add_marker(mark)

        with tempfile.NamedTemporaryFile() as tmp:
            tmp.write(value)
            tmp.flush()

            store.put_file(key, open(tmp.name, "rb"))

            assert store.get(key) == value

    def test_get_into_file(self, store, key, value, tmp_path):
        store.put(key, value)
        out_filename = os.path.join(str(tmp_path), "output")

        store.get_file(key, out_filename)

        assert open(out_filename, "rb").read() == value

    def test_get_into_stream(self, store, key, value):
        store.put(key, value)

        output = BytesIO()

        store.get_file(key, output)
        assert output.getvalue() == value

    def test_put_return_value(self, store, key, value):
        assert key == store.put(key, value)

    def test_put_file_return_value(self, store, key, value):
        assert key == store.put_file(key, BytesIO(value))

    def test_put_filename_return_value(self, store, key, value, request):
        if is_emulated_gcstore_test(store):
            mark = pytest.mark.xfail(
                reason="Triggers resumable upload, which isn't currently supported by the GC Emulator"
            )
            request.node.add_marker(mark)

        tmp = tempfile.NamedTemporaryFile(delete=False)
        try:
            tmp.write(value)
            tmp.close()

            assert key == store.put_file(key, tmp.name)
        finally:
            if os.path.exists(tmp.name):
                os.unlink(tmp.name)

    def test_delete(self, store, key, value):
        store.put(key, value)

        assert value == store.get(key)

        store.delete(key)

        with pytest.raises(KeyError):
            store.get(key)

    def test_multiple_delete_fails_without_error(self, store, key, value):
        store.put(key, value)

        store.delete(key)
        store.delete(key)
        store.delete(key)

    def test_can_delete_key_that_never_exists(self, store, key):
        store.delete(key)

    def test_key_iterator(self, store, key, key2, value, value2):
        store.put(key, value)
        store.put(key2, value2)

        keys = []
        for k in store.iter_keys():
            assert isinstance(k, str)
            keys.append(k)

        keys.sort()

        assert keys == sorted([key, key2])

    def test_key_iterator_with_prefix(self, store, key, key2, value):
        prefix = key
        key_prefix_1 = prefix + "_key1"
        key_prefix_2 = prefix + "_key2"
        store.put(key_prefix_1, value)
        store.put(key_prefix_2, value)
        store.put(key2, value)

        key_prefixes = []
        for k in store.iter_keys():
            key_prefixes.append(k)
        key_prefixes.sort()

        assert key_prefixes == sorted([key_prefix_1, key_prefix_2, key2])

        key_prefixes = []
        for k in store.iter_keys(prefix):
            key_prefixes.append(k)
        key_prefixes.sort()
        assert key_prefixes == sorted([key_prefix_1, key_prefix_2])

    def test_prefix_iterator(self, store, value):
        for k in [
            "X",
            "a1Xb1",
            "a1Xb1",
            "a2X",
            "a2Xb1",
            "a3",
            "a4Xb1Xc1",
            "a4Xb1Xc2",
            "a4Xb2Xc1",
            "a4Xb3",
        ]:
            store.put(k, value)

        prefixes = sorted(store.iter_prefixes("X"))
        assert prefixes == ["X", "a1X", "a2X", "a3", "a4X"]

        prefixes = sorted(store.iter_prefixes("X", prefix="a4X"))
        assert prefixes == ["a4Xb1X", "a4Xb2X", "a4Xb3"]

        prefixes = sorted(store.iter_prefixes("X", prefix="foo"))
        assert prefixes == []

    def test_keys(self, store, key, key2, value, value2):
        store.put(key, value)
        store.put(key2, value2)

        keys = store.keys()
        assert isinstance(keys, list)
        for k in keys:
            assert isinstance(k, str)

        assert set(keys) == {key, key2}

    def test_keys_with_prefix(self, store, key, key2, value):
        prefix = key
        key_prefix_1 = prefix + "_key1"
        key_prefix_2 = prefix + "_key2"
        store.put(key_prefix_1, value)
        store.put(key_prefix_2, value)
        store.put(key2, value)

        keys = store.keys()
        assert isinstance(keys, list)
        assert set(keys) == {key_prefix_1, key_prefix_2, key2}

        keys = store.keys(prefix)
        assert isinstance(keys, list)
        assert set(keys) == {key_prefix_1, key_prefix_2}

    def test_has_key(self, store, key, key2, value):
        store.put(key, value)

        assert key in store
        assert key2 not in store

    def test_has_key_with_delete(self, store, key, value):
        assert key not in store

        store.put(key, value)
        assert key in store

        store.delete(key)
        assert key not in store

        store.put(key, value)
        assert key in store

    def test_get_with_delete(self, store, key, value):
        with pytest.raises(KeyError):
            store.get(key)

        store.put(key, value)
        store.get(key)

        store.delete(key)

        with pytest.raises(KeyError):
            store.get(key)

        store.put(key, value)
        store.get(key)

    def test_max_key_length(self, store, max_key, value):
        new_key = store.put(max_key, value)

        assert new_key == max_key
        assert value == store.get(max_key)

    def test_a_lot_of_puts(self, store, key, value):
        a_lot = 20

        for i in range(a_lot):
            key = key + f"_{i}"
            store.put(key, value)

    # We should expand this to include more tests interfacing with other
    # FileSystem APIs like ParquetFile.
    def test_parquet_file(self, store):
        # Skip if were using a SQLAlchemyStore
        from minimalkv.db.sql import SQLAlchemyStore

        if isinstance(store, SQLAlchemyStore):
            pytest.skip("SQLAlchemyStore doesn't support ParquetFile yet")
        with open("tests/test.parquet", "rb") as f:
            store.put_file("test.parquet", f)
        # Open parquet file
        f = store.open("test.parquet")
        p = ParquetFile(f)
        assert p.metadata.num_columns == 13
        # Read metadata from parquet file


# small extra time added to account for variance
TTL_MARGIN = 1


class TTLStore:
    @pytest.fixture
    def ustore(self, store):
        return UUIDDecorator(store)

    @pytest.fixture(params=["hash", "uuid", "hmac", "prefix"])
    def dstore(self, request, store, secret_key, ustore):
        if request.param == "hash":
            return HashDecorator(store)
        elif request.param == "uuid":
            return ustore
        elif request.param == "hmac":
            return HMACDecorator(secret_key, store)
        elif request.param == "prefix":
            return PrefixDecorator("SaMpLe_PrEfIX", store)

    @pytest.fixture(params=[0.4, 1])
    def small_ttl(self, request):
        return request.param

    def test_put_with_negative_ttl_throws_error(self, store, key, value):
        with pytest.raises(ValueError):
            store.put(key, value, ttl_secs=-1)

    def test_put_with_non_numeric_ttl_throws_error(self, store, key, value):
        with pytest.raises(ValueError):
            store.put(key, value, ttl_secs="badttl")

    def test_put_with_ttl_argument(self, store, key, value, small_ttl):
        store.put(key, value, small_ttl)

        time.sleep(small_ttl + TTL_MARGIN)
        with pytest.raises(KeyError):
            store.get(key)

    def test_put_set_default(self, store, key, value, small_ttl):
        store.default_ttl_secs = small_ttl

        store.put(key, value)

        time.sleep(small_ttl + TTL_MARGIN)
        with pytest.raises(KeyError):
            store.get(key)

    def test_put_file_with_ttl_argument(self, store, key, value, small_ttl):
        store.put_file(key, BytesIO(value), small_ttl)

        time.sleep(small_ttl + TTL_MARGIN)
        with pytest.raises(KeyError):
            store.get(key)

    def test_put_file_set_default(self, store, key, value, small_ttl):
        store.default_ttl_secs = small_ttl

        store.put_file(key, BytesIO(value))

        time.sleep(small_ttl + TTL_MARGIN)
        with pytest.raises(KeyError):
            store.get(key)

    def test_uuid_decorator(self, ustore, value):
        key = ustore.put(None, value)

        assert key

    def test_advertises_ttl_features(self, store):
        assert store.ttl_support is True
        assert hasattr(store, "ttl_support")
        assert getattr(store, "ttl_support") is True

    def test_advertises_ttl_features_through_decorator(self, dstore):
        assert dstore.ttl_support is True
        assert hasattr(dstore, "ttl_support")
        assert getattr(dstore, "ttl_support") is True

    def test_can_pass_ttl_through_decorator(self, dstore, key, value):
        dstore.put(key, value, ttl_secs=10)


class OpenSeekTellStore:
    def test_open_seek_and_tell_empty_value(self, store, key):
        value = b""
        store.put(key, value)
        ok = store.open(key)
        assert ok.seekable()
        assert ok.seek(10) == 10
        assert ok.tell() == 10
        assert ok.seek(-6, 1) == 4
        assert ok.tell() == 4
        with pytest.raises(IOError):
            ok.seek(-1, 0)
        with pytest.raises(IOError):
            ok.seek(-6, 1)
        with pytest.raises(IOError):
            ok.seek(-1, 2)

        assert ok.tell() == 4
        assert b"" == ok.read(1)

    def test_open_seek_and_tell(self, store, key, long_value):
        store.put(key, long_value)
        ok = store.open(key)
        assert ok.seekable()
        assert ok.readable()
        ok.seek(10)
        assert ok.tell() == 10
        ok.seek(-6, 1)
        assert ok.tell() == 4
        with pytest.raises(IOError):
            ok.seek(-1, 0)
        with pytest.raises(IOError):
            ok.seek(-6, 1)
        with pytest.raises(IOError):
            ok.seek(-len(long_value) - 1, 2)

        assert ok.tell() == 4
        assert long_value[4:5] == ok.read(1)
        assert ok.tell() == 5
        ok.seek(-1, 2)
        length_lv = len(long_value)
        assert long_value[length_lv - 1 : length_lv] == ok.read(1)
        assert ok.tell() == length_lv
        ok.seek(length_lv + 10, 0)
        assert ok.tell() == length_lv + 10
        assert b"" == ok.read()

        ok.close()
        with pytest.raises(ValueError):
            ok.tell()
        with pytest.raises(ValueError):
            ok.read(1)
        with pytest.raises(ValueError):
            ok.seek(10)



import base64
import json
import pathlib

import pytest

from minimalkv._get_store import get_store_from_url
from minimalkv._store_creation import create_store
from minimalkv._urls import url2dict
from minimalkv.net.gcstore import GoogleCloudStore

storage = pytest.importorskip("google.cloud.storage")
from google.auth.credentials import AnonymousCredentials
from google.auth.exceptions import RefreshError


def test_create_store_gcstore(mocker):
    mock_hgcstore = mocker.patch("minimalkv._hstores.HGoogleCloudStore")
    mock_gcstore = mocker.patch("minimalkv.net.gcstore.GoogleCloudStore")

    anon_credentials = AnonymousCredentials()
    create_store(
        "gcs",
        {
            "credentials": anon_credentials,
            "bucket_name": "test_bucket",
            "create_if_missing": True,
            "bucket_creation_location": "EUROPE-WEST1",
        },
    )
    mock_gcstore.assert_called_once_with(
        credentials=anon_credentials,
        bucket_name="test_bucket",
        create_if_missing=True,
        bucket_creation_location="EUROPE-WEST1",
    )
    mock_hgcstore.assert_not_called()


def test_create_store_hgcstore(mocker):
    mock_hgcstore = mocker.patch("minimalkv._hstores.HGoogleCloudStore")

    anon_credentials = AnonymousCredentials()
    create_store(
        "hgcs",
        {
            "credentials": anon_credentials,
            "bucket_name": "test_bucket",
            "create_if_missing": True,
            "bucket_creation_location": "EUROPE-WEST1",
        },
    )
    mock_hgcstore.assert_called_once_with(
        credentials=anon_credentials,
        bucket_name="test_bucket",
        create_if_missing=True,
        bucket_creation_location="EUROPE-WEST1",
    )


SHORT_URL = (
    f"gcs://{base64.urlsafe_b64encode(b'some bytes+/=asdf').decode()}"
    f"@bucket_name?create_if_missing=true&bucket_creation_location=WESTINDIES",
    {
        "type": "gcs",
        "credentials": b"some bytes+/=asdf",
        "bucket_name": "bucket_name",
        "create_if_missing": True,
        "bucket_creation_location": "WESTINDIES",
    },
)
ACTUAL_URL = (
    f"gcs://{base64.urlsafe_b64encode(pathlib.Path('tests/storefact/gcstore_cred_example.json').read_bytes()).decode()}"
    f"@default_bucket?create_if_missing=false",
    {
        "type": "gcs",
        "credentials": pathlib.Path(
            "tests/storefact/gcstore_cred_example.json"
        ).read_bytes(),
        "bucket_name": "default_bucket",
        "create_if_missing": False,
    },
)


@pytest.mark.parametrize("url, expected", [SHORT_URL, ACTUAL_URL])
def test_url2dict(url, expected):
    assert url2dict(url) == expected


def test_json_decode():
    url, _ = ACTUAL_URL
    creds = url2dict(url)["credentials"]
    with open("tests/storefact/gcstore_cred_example.json") as file:
        assert json.loads(creds) == json.load(file)


def test_complete():
    url, expected = ACTUAL_URL
    store = get_store_from_url(url)
    assert isinstance(store, GoogleCloudStore)
    assert store.bucket_name == expected["bucket_name"]  # type: ignore
    assert store.project_name == "central-splice-296415"  # type: ignore
    with pytest.raises(RefreshError):
        store.get("somekey")



import pytest


class UrlStore:
    def test_url_for_for_generates_url_for(self, store, key, value):
        store.put(key, value)
        assert store.url_for(key)

    def test_url_for_generation_does_not_check_exists(self, store, key):
        store.url_for(key)

    def test_url_for_generation_checks_valid_key(self, store, invalid_key):
        with pytest.raises(ValueError):
            store.url_for(invalid_key)



from io import BytesIO
from typing import IO, Iterable, Iterator, List, Optional, Union

from minimalkv._constants import VALID_KEY_RE
from minimalkv._mixins import UrlMixin

# Only here to keep backwards-compatability
key_type = str


class KeyValueStore:
    """
    Class to access a key-value store.

    Supported keys are ascii-strings containing alphanumeric characters or symbols out
    of ``minimalkv._constants.VALID_NON_NUM`` of length not greater than 250. Values
    (or records) are stored as raw bytes.
    """

    def __contains__(self, key: str) -> bool:
        """Check if the store has an entry at key.

        Parameters
        ----------
        key : str
            The key whose existence should be verified.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was an error accessing the store.
        """
        self._check_valid_key(key)
        return self._has_key(key)

    def __iter__(self) -> Iterable[str]:
        """Iterate over all keys in the store.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        return self.iter_keys()

    def delete(self, key: str) -> Optional[str]:
        """Delete data at key.

        Does not raise an error if the key does not exist.

        Parameters
        ----------
        key: str
            The key of data to be deleted.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was an error deleting.
        """
        self._check_valid_key(key)
        return self._delete(key)

    def get(self, key: str) -> bytes:
        """Return data at key as a bytestring.

        Parameters
        ----------
        key : str
            The key to be read.

        Returns
        -------
        data : str
            Value associated with the key as a ``bytes`` object.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If the file could not be read.
        KeyError
            If the key was not found.
        """
        self._check_valid_key(key)
        return self._get(key)

    def get_file(self, key: str, file: Union[str, IO]) -> str:
        """Write data at key to file.

        Like :meth:`~mininmalkv.KeyValueStore.put_file`, this method allows backends to
        implement a specialized function if data needs to be written to disk or streamed.

        If ``file`` is a string, contents of ``key`` are written to a newly created file
        with the filename ``file``. Otherwise the data will be written using the
        ``write`` method of ``file``.

        Parameters
        ----------
        key : str
            The key to be read.
        file : file-like or str
            Output filename or file-like object with a ``write`` method.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was a problem reading or writing data.
        KeyError
            If the key was not found.
        """
        self._check_valid_key(key)
        if isinstance(file, str):
            return self._get_filename(key, file)
        else:
            return self._get_file(key, file)

    def iter_keys(self, prefix: str = "") -> Iterator[str]:
        """Iterate over all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only iterate over keys starting with prefix. Iterate over all keys if empty.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        raise NotImplementedError

    def iter_prefixes(self, delimiter: str, prefix: str = "") -> Iterator[str]:
        """
        Iterate over unique prefixes in the store up to delimiter, starting with prefix.

        If ``prefix`` contains ``delimiter``, return the prefix up to the first
        occurence of delimiter after the prefix.

        The default uses an naive key iteration. Some backends may implement more
        efficient methods.

        Parameters
        ----------
        delimiter : str, optional, default = ''
            Delimiter up to which to iterate over prefixes.
        prefix : str, optional, default = ''
            Only iterate over prefixes starting with prefix.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        dlen = len(delimiter)
        plen = len(prefix)
        memory = set()

        for k in self.iter_keys(prefix):
            pos = k.find(delimiter, plen)
            if pos >= 0:
                k = k[: pos + dlen]

            if k not in memory:
                yield k
                memory.add(k)

    def keys(self, prefix: str = "") -> List[str]:
        """List all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only list keys starting with prefix. List all keys if empty.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        return list(self.iter_keys(prefix))

    def open(self, key: str) -> IO:
        """Open record at key.

        Parameters
        ----------
        key : str
            Key to open.

        Returns
        -------
        file: file-like
            Read-only file-like object for reading data at key.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If the file could not be read.
        KeyError
            If the key was not found.
        """
        self._check_valid_key(key)
        return self._open(key)

    def put(self, key: str, data: bytes) -> str:
        """Store bytestring data at key.

        Parameters
        ----------
        key : str
            The key under which the data is to be stored.
        data : bytes
            Data to be stored at key, must be of type  ``bytes``.

        Returns
        -------
        str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If storing failed or the file could not be read.
        """
        self._check_valid_key(key)
        if not isinstance(data, bytes):
            raise OSError("Provided data is not of type bytes")
        return self._put(key, data)

    def put_file(self, key: str, file: Union[str, IO]) -> str:
        """Store contents of file at key.

        Store data from a file into key. ``file`` can be a string, which will be
        interpreted as a filename, or an object with a ``read()`` method.

        If ``file`` is a filename, the file might be removed while storing to avoid
        unnecessary copies. To prevent this, pass the opened file instead.

        Parameters
        ----------
        key : str
            Key where to store data in file.
        file : file-like or str
            A filename or a file-like object with a read method.

        Returns
        -------
        key: str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was a problem moving the file in.

        """
        self._check_valid_key(key)
        if isinstance(file, str):
            return self._put_filename(key, file)
        else:
            return self._put_file(key, file)

    def _check_valid_key(self, key: str) -> None:
        """Check if a key is valid and raise a ValueError if it is not.

        Always use this method to check whether a key is valid.

        Parameters
        ----------
        key : str
            The key to be checked.

        Raises
        ------
        ValueError
            If the key is not valid.
        """
        if not isinstance(key, key_type):
            raise ValueError(f"The key {key} is not a valid key type.")
        if not VALID_KEY_RE.match(key):
            raise ValueError(f"The key {key} contains illegal characters.")

    def _delete(self, key: str):
        """Delete the data at key in store."""
        raise NotImplementedError

    def _get(self, key: str) -> bytes:
        """Read data at key in store.

        Parameters
        ----------
        key : str
            Key of value to be retrieved.
        """
        buf = BytesIO()

        self._get_file(key, buf)

        return buf.getvalue()

    def _get_file(self, key: str, file: IO) -> str:
        """Write data at key to file-like object file.

        Parameters
        ----------
        key : str
            Key of data to be written to file.
        file : file-like
            File-like object with a *write* method to be written.
        """
        bufsize = 1024 * 1024

        # note: we do not use a context manager here or close the source.
        # the source goes out of scope shortly after, taking care of the issue
        # this allows us to support file-like objects without close as well,
        # such as BytesIO.
        source = self.open(key)
        try:
            while True:
                buf = source.read(bufsize)
                file.write(buf)

                if len(buf) < bufsize:
                    break
        finally:
            source.close()

        return key

    def _get_filename(self, key: str, filename: str) -> str:
        """Write data at key to file at filename.

        Parameters
        ----------
        key : str
            Key of data to be written to file at filename.
        filename : str
            Name of file to be written.
        """
        with open(filename, "wb") as dest:
            return self._get_file(key, dest)

    def _has_key(self, key: str) -> bool:
        """Check the existence of key in store.

        Parameters
        ----------
        key : str
            Key to check the existance of.
        """
        return key in self.keys()

    def _open(self, key: str) -> IO:
        """Open record at key.

        Parameters
        ----------
        key : str
            Key of record to open.

        Returns
        -------
        file: file-like
            Opened file.
        """
        raise NotImplementedError

    def _put(self, key: str, data: bytes) -> str:
        """Store bytestring data at key.

        Parameters
        ----------
        key : str
            Key under which data should be stored.
        data : bytes
            Data to be stored.

        Returns
        -------
        key : str
            Key where data was stored.

        """
        return self._put_file(key, BytesIO(data))

    def _put_file(self, key: str, file: IO) -> str:
        """Store data from file-like object at key.

        Parameters
        ----------
        key : str
            Key at which to store contents of file.
        file : file-like
            File-like object to store data from.

        Returns
        -------
        key : str
            Key where data was stored.

        """
        raise NotImplementedError

    def _put_filename(self, key: str, filename: str) -> str:
        """Store data from file at filename at key.

        Parameters
        ----------
        key : str
            Key under which data should be stored.
        filename : str
            Filename of file to store.

        Returns
        -------
        key: str
            Key where data was stored.

        """
        with open(filename, "rb") as source:
            return self._put_file(key, source)


class UrlKeyValueStore(UrlMixin, KeyValueStore):
    """Class is deprecated. Use the :class:`.UrlMixin` instead.

    .. deprecated:: 0.9

    """

    pass



from unittest import mock

optional_dependencies = [
    "azure",
    "boto",
    "dulwich",
    "fsspec",
    "gcsfs",
    "google",
    "redis",
    "sqlalchemy",
]

orig_import = __import__


def mock_optional_dependencies(name, *args):
    if name in optional_dependencies:
        raise ImportError(f"Cannot import optional dependency {name}.")

    return orig_import(name, *args)


def test_import():
    with mock.patch("builtins.__import__", side_effect=mock_optional_dependencies):
        from minimalkv import get_store_from_url

        get_store_from_url("hfs:///tmp")
